@@ -1,23 +1,29 @@
-use crate::ast::Identifier;
 use crate::ast::{Expr, Stmt};
-use crate::calculus;
-use crate::kalk_num::KalkNum;
+use crate::ast::{Identifier, RangedVar};
+use crate::calculation_result::CalculationResult;
+use crate::errors::KalkError;
+use crate::kalk_value::KalkValue;
 use crate::lexer::TokenKind;
-use crate::parser::CalcError;
 use crate::parser::DECL_UNIT;
-use crate::prelude;
 use crate::symbol_table::SymbolTable;
+use crate::{as_number_or_zero, numerical};
+use crate::{float, prelude};
+
+pub const DEFAULT_MAX_RECURSION_DEPTH: u32 = 256;
 
 pub struct Context<'a> {
     pub symbol_table: &'a mut SymbolTable,
     angle_unit: String,
     #[cfg(feature = "rug")]
     precision: u32,
-    sum_n_value: Option<i128>,
+    sum_variables: Option<Vec<SumVar>>,
     #[cfg(not(target_arch = "wasm32"))]
     timeout: Option<u128>,
     #[cfg(not(target_arch = "wasm32"))]
     start_time: std::time::SystemTime,
+    is_approximation: bool,
+    recursion_depth: u32,
+    max_recursion_depth: u32,
 }
 
 impl<'a> Context<'a> {
@@ -32,24 +38,36 @@ impl<'a> Context<'a> {
             symbol_table,
             #[cfg(feature = "rug")]
             precision,
-            sum_n_value: None,
+            sum_variables: None,
             #[cfg(not(target_arch = "wasm32"))]
-            timeout: timeout,
+            timeout,
             #[cfg(not(target_arch = "wasm32"))]
             start_time: std::time::SystemTime::now(),
+            is_approximation: false,
+            recursion_depth: 0,
+            max_recursion_depth: DEFAULT_MAX_RECURSION_DEPTH,
         }
     }
 
-    pub fn interpret(&mut self, statements: Vec<Stmt>) -> Result<Option<KalkNum>, CalcError> {
+    pub fn set_max_recursion_depth(mut self, depth: u32) -> Self {
+        self.max_recursion_depth = depth;
+
+        self
+    }
+
+    pub fn interpret(
+        &mut self,
+        statements: Vec<Stmt>,
+    ) -> Result<Option<CalculationResult>, KalkError> {
         for (i, stmt) in statements.iter().enumerate() {
             let num = eval_stmt(self, stmt)?;
 
             // Insert the last value into the `ans` variable.
-            self.symbol_table.set(if (&num.unit).len() > 0 {
+            self.symbol_table.set(if num.has_unit() {
                 Stmt::VarDecl(
                     Identifier::from_full_name("ans"),
                     Box::new(Expr::Unit(
-                        num.unit.clone(),
+                        num.get_unit().unwrap().to_string(),
                         Box::new(crate::ast::build_literal_ast(&num)),
                     )),
                 )
@@ -62,7 +80,7 @@ impl<'a> Context<'a> {
 
             if i == statements.len() - 1 {
                 if let Stmt::Expr(_) = stmt {
-                    return Ok(Some(num));
+                    return Ok(Some(CalculationResult::new(num, 10, self.is_approximation)));
                 }
             }
         }
@@ -71,55 +89,71 @@ impl<'a> Context<'a> {
     }
 }
 
-fn eval_stmt(context: &mut Context, stmt: &Stmt) -> Result<KalkNum, CalcError> {
+struct SumVar {
+    name: String,
+    value: i128,
+}
+
+fn eval_stmt(context: &mut Context, stmt: &Stmt) -> Result<KalkValue, KalkError> {
     match stmt {
         Stmt::VarDecl(_, _) => eval_var_decl_stmt(context, stmt),
         Stmt::FnDecl(_, _, _) => eval_fn_decl_stmt(),
         Stmt::UnitDecl(_, _, _) => eval_unit_decl_stmt(),
-        Stmt::Expr(expr) => eval_expr_stmt(context, &expr),
+        Stmt::Expr(expr) => eval_expr_stmt(context, expr),
     }
 }
 
-fn eval_var_decl_stmt(context: &mut Context, stmt: &Stmt) -> Result<KalkNum, CalcError> {
+fn eval_var_decl_stmt(context: &mut Context, stmt: &Stmt) -> Result<KalkValue, KalkError> {
     context.symbol_table.insert(stmt.clone());
-    Ok(KalkNum::from(1))
+    Ok(KalkValue::from(1))
 }
 
-fn eval_fn_decl_stmt() -> Result<KalkNum, CalcError> {
-    Ok(KalkNum::from(1)) // Nothing needs to happen here, since the parser will already have added the FnDecl's to the symbol table.
+fn eval_fn_decl_stmt() -> Result<KalkValue, KalkError> {
+    Ok(KalkValue::from(1)) // Nothing needs to happen here, since the parser will already have added the FnDecl's to the symbol table.
 }
 
-fn eval_unit_decl_stmt() -> Result<KalkNum, CalcError> {
-    Ok(KalkNum::from(1))
+fn eval_unit_decl_stmt() -> Result<KalkValue, KalkError> {
+    Ok(KalkValue::from(1))
 }
 
-fn eval_expr_stmt(context: &mut Context, expr: &Expr) -> Result<KalkNum, CalcError> {
-    eval_expr(context, &expr, "")
+fn eval_expr_stmt(context: &mut Context, expr: &Expr) -> Result<KalkValue, KalkError> {
+    eval_expr(context, expr, None)
 }
 
 pub(crate) fn eval_expr(
     context: &mut Context,
     expr: &Expr,
-    unit: &str,
-) -> Result<KalkNum, CalcError> {
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
     #[cfg(not(target_arch = "wasm32"))]
     if let (Ok(elapsed), Some(timeout)) = (context.start_time.elapsed(), context.timeout) {
         if elapsed.as_millis() >= timeout {
-            return Err(CalcError::TimedOut);
+            return Err(KalkError::TimedOut);
         }
     }
 
     match expr {
-        Expr::Binary(left, op, right) => eval_binary_expr(context, &left, op, &right, unit),
+        Expr::Binary(left, op, right) => eval_binary_expr(context, left, op, right, unit),
         Expr::Unary(op, expr) => eval_unary_expr(context, op, expr, unit),
         Expr::Unit(identifier, expr) => eval_unit_expr(context, identifier, expr),
         Expr::Var(identifier) => eval_var_expr(context, identifier, unit),
         Expr::Literal(value) => eval_literal_expr(context, *value, unit),
-        Expr::Group(expr) => eval_group_expr(context, &expr, unit),
+        Expr::Boolean(value) => Ok(KalkValue::Boolean(*value)),
+        Expr::Group(expr) => eval_group_expr(context, expr, unit),
         Expr::FnCall(identifier, expressions) => {
-            eval_fn_call_expr(context, identifier, expressions, unit)
+            context.recursion_depth += 1;
+            let res = eval_fn_call_expr(context, identifier, expressions, unit);
+            context.recursion_depth -= 1;
+            res
         }
         Expr::Piecewise(pieces) => eval_piecewise(context, pieces, unit),
+        Expr::Vector(values) => eval_vector(context, values),
+        Expr::Matrix(rows) => eval_matrix(context, rows),
+        Expr::Indexer(var, indexes) => eval_indexer(context, var, indexes, unit),
+        Expr::Comprehension(left, conditions, vars) => Ok(KalkValue::Vector(eval_comprehension(
+            context, left, conditions, vars,
+        )?)),
+        Expr::Equation(left, right, identifier) => eval_equation(context, left, right, identifier),
     }
 }
 
@@ -128,28 +162,33 @@ fn eval_binary_expr(
     left_expr: &Expr,
     op: &TokenKind,
     right_expr: &Expr,
-    unit: &str,
-) -> Result<KalkNum, CalcError> {
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
     if let TokenKind::ToKeyword = op {
         // TODO: When the unit conversion function takes a Float instead of Expr,
         // move this to the match statement further down.
         if let Expr::Var(right_unit) = right_expr {
-            let left_unit = eval_expr(context, left_expr, "")?.unit;
-            return convert_unit(context, left_expr, &left_unit, &right_unit.full_name);
+            let left_unit = eval_expr(context, left_expr, None)?.get_unit().cloned();
+            return convert_unit(
+                context,
+                left_expr,
+                left_unit.as_ref(),
+                Some(&right_unit.full_name),
+            );
             // TODO: Avoid evaluating this twice.
         }
     }
 
-    let left = eval_expr(context, left_expr, "")?;
-    let mut right = eval_expr(context, right_expr, "")?;
+    let left = eval_expr(context, left_expr, None)?;
+    let mut right = eval_expr(context, right_expr, None)?;
     if let Expr::Unary(TokenKind::Percent, _) = right_expr {
-        right = right.mul(context, left.clone());
+        right = right.mul(context, left.clone())?;
         if let TokenKind::Star = op {
             return Ok(right);
         }
     }
 
-    let mut result = match op {
+    let result = match op {
         TokenKind::Plus => left.add(context, right),
         TokenKind::Minus => left.sub(context, right),
         TokenKind::Star => left.mul(context, right),
@@ -162,11 +201,15 @@ fn eval_binary_expr(
         TokenKind::LessThan => left.less_than(context, right),
         TokenKind::GreaterOrEquals => left.greater_or_equals(context, right),
         TokenKind::LessOrEquals => left.less_or_equals(context, right),
-        _ => KalkNum::from(1),
-    };
+        TokenKind::And => left.and(&right),
+        TokenKind::Or => left.or(&right),
+        _ => Ok(KalkValue::from(1f64)),
+    }?;
 
-    if unit.len() > 0 {
-        result.unit = unit.to_string();
+    if unit.is_some() {
+        if let KalkValue::Number(real, imaginary, _) = result {
+            return Ok(KalkValue::Number(real, imaginary, unit.cloned()));
+        }
     };
 
     Ok(result)
@@ -176,15 +219,19 @@ fn eval_unary_expr(
     context: &mut Context,
     op: &TokenKind,
     expr: &Expr,
-    unit: &str,
-) -> Result<KalkNum, CalcError> {
-    let num = eval_expr(context, &expr, unit)?;
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
+    let num = eval_expr(context, expr, unit)?;
 
     match op {
-        TokenKind::Minus => Ok(num.mul(context, KalkNum::from(-1f64))),
-        TokenKind::Percent => Ok(num.mul(context, KalkNum::from(0.01f64))),
-        TokenKind::Exclamation => Ok(prelude::special_funcs::factorial(num)),
-        _ => Err(CalcError::InvalidOperator),
+        TokenKind::Minus => num.mul(context, KalkValue::from(-1f64)),
+        TokenKind::Not => match num {
+            KalkValue::Boolean(boolean) => Ok(KalkValue::Boolean(!boolean)),
+            _ => Err(KalkError::InvalidOperator),
+        },
+        TokenKind::Percent => num.mul(context, KalkValue::from(0.01f64)),
+        TokenKind::Exclamation => prelude::special_funcs::factorial(num),
+        _ => Err(KalkError::InvalidOperator),
     }
 }
 
@@ -192,21 +239,32 @@ fn eval_unit_expr(
     context: &mut Context,
     identifier: &str,
     expr: &Expr,
-) -> Result<KalkNum, CalcError> {
+) -> Result<KalkValue, KalkError> {
     let angle_unit = &context.angle_unit.clone();
     if (identifier == "rad" || identifier == "deg") && angle_unit != identifier {
-        return convert_unit(context, expr, identifier, angle_unit);
+        return convert_unit(
+            context,
+            expr,
+            Some(&identifier.to_string()),
+            Some(angle_unit),
+        );
     }
 
-    eval_expr(context, expr, identifier)
+    eval_expr(context, expr, Some(&identifier.to_string()))
 }
 
 pub fn convert_unit(
     context: &mut Context,
     expr: &Expr,
-    from_unit: &str,
-    to_unit: &str,
-) -> Result<KalkNum, CalcError> {
+    from_unit: Option<&String>,
+    to_unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
+    let (from_unit, to_unit) = if let (Some(from_unit), Some(to_unit)) = (from_unit, to_unit) {
+        (from_unit, to_unit)
+    } else {
+        return Err(KalkError::InvalidUnit);
+    };
+
     if let Some(Stmt::UnitDecl(_, _, unit_def)) =
         context.symbol_table.get_unit(to_unit, from_unit).cloned()
     {
@@ -215,30 +273,29 @@ pub fn convert_unit(
             Box::new(expr.clone()),
         ));
 
-        let num = eval_expr(context, &unit_def, "")?;
-        Ok(KalkNum::new_with_imaginary(
-            num.value,
-            to_unit.into(),
-            num.imaginary_value,
-        ))
+        let (real, imaginary, _) = as_number_or_zero!(eval_expr(context, &unit_def, None)?);
+        Ok(KalkValue::Number(real, imaginary, Some(to_unit.clone())))
     } else {
-        Err(CalcError::InvalidUnit)
+        Err(KalkError::InvalidUnit)
     }
 }
 
 fn eval_var_expr(
     context: &mut Context,
     identifier: &Identifier,
-    unit: &str,
-) -> Result<KalkNum, CalcError> {
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
     // If there is a constant with this name, return a literal expression with its value
     if let Some(value) = prelude::CONSTANTS.get(identifier.full_name.as_ref() as &str) {
         return eval_expr(context, &Expr::Literal(*value), unit);
     }
 
-    if identifier.full_name == "n" {
-        if let Some(value) = context.sum_n_value {
-            return Ok(KalkNum::from(value));
+    if let Some(sum_variables) = &context.sum_variables {
+        let sum_variable = sum_variables
+            .iter()
+            .find(|x| x.name == identifier.full_name);
+        if let Some(sum_variable) = sum_variable {
+            return Ok(KalkValue::from(sum_variable.value));
         }
     }
 
@@ -247,24 +304,41 @@ fn eval_var_expr(
         .symbol_table
         .get_var(identifier.full_name.as_ref() as &str)
         .cloned();
-    match var_decl {
-        Some(Stmt::VarDecl(_, expr)) => eval_expr(context, &expr, unit),
-        _ => Err(CalcError::UndefinedVar(identifier.full_name.clone())),
+    if let Some(Stmt::VarDecl(_, expr)) = var_decl {
+        eval_expr(context, &expr, unit)
+    } else {
+        Err(KalkError::UndefinedVar(identifier.full_name.clone()))
     }
 }
 
 #[allow(unused_variables)]
-fn eval_literal_expr(context: &mut Context, value: f64, unit: &str) -> Result<KalkNum, CalcError> {
-    let mut num: KalkNum = value.into();
-    num.unit = unit.into();
+#[cfg(feature = "rug")]
+fn eval_literal_expr(
+    context: &mut Context,
+    value: f64,
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
+    let mut float = float!(value);
+    float.set_prec(context.precision);
 
-    #[cfg(feature = "rug")]
-    num.value.set_prec(context.precision);
-
-    Ok(num)
+    Ok(KalkValue::Number(float, float!(0), unit.cloned()))
 }
 
-fn eval_group_expr(context: &mut Context, expr: &Expr, unit: &str) -> Result<KalkNum, CalcError> {
+#[allow(unused_variables)]
+#[cfg(not(feature = "rug"))]
+fn eval_literal_expr(
+    context: &mut Context,
+    value: f64,
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
+    Ok(KalkValue::Number(float!(value), float!(0), unit.cloned()))
+}
+
+fn eval_group_expr(
+    context: &mut Context,
+    expr: &Expr,
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
     eval_expr(context, expr, unit)
 }
 
@@ -272,14 +346,100 @@ pub(crate) fn eval_fn_call_expr(
     context: &mut Context,
     identifier: &Identifier,
     expressions: &[Expr],
-    unit: &str,
-) -> Result<KalkNum, CalcError> {
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
+    if context.recursion_depth > context.max_recursion_depth {
+        return Err(KalkError::StackOverflow);
+    }
+
+    if identifier.prime_count > 0 {
+        context.is_approximation = true;
+    }
+
+    // Special functions
+    match identifier.full_name.as_ref() {
+        "sum" | "prod" => {
+            // If loop, eg. sum(k=1, 100, k)
+            if let Expr::Binary(left, TokenKind::Equals, right) = &expressions[0] {
+                if let Expr::Var(var_identifier) = &**left {
+                    // Make sure exactly 3 arguments were supplied.
+                    if expressions.len() != 3 {
+                        return Err(KalkError::IncorrectAmountOfArguments(
+                            3,
+                            "sum/prod".into(),
+                            expressions.len(),
+                        ));
+                    }
+
+                    //(var_identifier.pure_name.as_ref(), &**right);
+                    return eval_loop(
+                        context,
+                        identifier,
+                        &var_identifier.pure_name,
+                        right,
+                        &expressions[1],
+                        &expressions[2],
+                        unit.cloned(),
+                    );
+                }
+            }
+        }
+        "integrate" => {
+            context.is_approximation = true;
+
+            return match expressions.len() {
+                3 => numerical::integrate_with_unknown_variable(
+                    context,
+                    &expressions[0],
+                    &expressions[1],
+                    &expressions[2],
+                ),
+                4 => numerical::integrate(
+                    context,
+                    &expressions[0],
+                    &expressions[1],
+                    &expressions[2],
+                    if let Expr::Var(integration_variable) = &expressions[3] {
+                        &integration_variable.full_name[1..]
+                    } else {
+                        return Err(KalkError::ExpectedDx);
+                    },
+                ),
+                _ => Err(KalkError::IncorrectAmountOfArguments(
+                    3,
+                    "integrate".into(),
+                    expressions.len(),
+                )),
+            };
+        }
+        _ => (),
+    }
+
+    // Prelude vector function
+    if prelude::is_vector_func(&identifier.full_name) {
+        let mut values = Vec::new();
+        for expression in expressions {
+            let value = eval_expr(context, expression, None)?;
+            if expressions.len() == 1 {
+                if let KalkValue::Vector(internal_values) = value {
+                    values = internal_values;
+                    break;
+                }
+            }
+
+            values.push(value);
+        }
+
+        return prelude::call_vector_func(&identifier.full_name, KalkValue::Vector(values))
+            .unwrap_or_else(|| Ok(KalkValue::nan()));
+    }
+
     // Prelude
     let prelude_func = match expressions.len() {
         1 => {
-            let x = eval_expr(context, &expressions[0], "")?;
+            let x = eval_expr(context, &expressions[0], None)?;
             if identifier.prime_count > 0 {
-                return calculus::derive_func(context, &identifier, x);
+                return numerical::derive_func(context, identifier, x);
             } else {
                 prelude::call_unary_func(
                     context,
@@ -290,8 +450,8 @@ pub(crate) fn eval_fn_call_expr(
             }
         }
         2 => {
-            let x = eval_expr(context, &expressions[0], "")?;
-            let y = eval_expr(context, &expressions[1], "")?;
+            let x = eval_expr(context, &expressions[0], None)?;
+            let y = eval_expr(context, &expressions[1], None)?;
             prelude::call_binary_func(
                 context,
                 &identifier.full_name,
@@ -304,84 +464,60 @@ pub(crate) fn eval_fn_call_expr(
     };
 
     if let Some((result, _)) = prelude_func {
-        return Ok(result);
-    }
+        if result.is_err() && expressions.len() == 1 {
+            let x = eval_expr(context, &expressions[0], None)?;
 
-    // Special functions
-    match identifier.full_name.as_ref() {
-        "sum" | "Σ" | "∑" | "prod" | "∏" => {
-            // Make sure exactly 3 arguments were supplied.
-            if expressions.len() != 3 {
-                return Err(CalcError::IncorrectAmountOfArguments(
-                    3,
-                    "sum/prod".into(),
-                    expressions.len(),
-                ));
-            }
+            // If a vector/matrix was given, call the function on every item
+            // in the vector/matrix.
+            if let KalkValue::Vector(values) = x {
+                let mut new_values = Vec::new();
+                let mut success = true;
+                for value in values {
+                    if let Some(result) = prelude::call_unary_func(
+                        context,
+                        &identifier.full_name,
+                        value,
+                        &context.angle_unit.clone(),
+                    ) {
+                        new_values.push(result.0?);
+                    } else {
+                        success = false;
+                        break;
+                    }
+                }
 
-            let start = eval_expr(context, &expressions[0], "")?.to_f64() as i128;
-            let end = eval_expr(context, &expressions[1], "")?.to_f64() as i128;
-            let sum_else_prod = match identifier.full_name.as_ref() {
-                "sum" => true,
-                "Σ" => true,
-                "∑" => true,
-                "prod" => false,
-                "∏" => false,
-                _ => unreachable!(),
-            };
-            let mut sum = if sum_else_prod {
-                KalkNum::default()
-            } else {
-                KalkNum::from(1f64)
-            };
+                if success {
+                    return Ok(KalkValue::Vector(new_values));
+                }
+            } else if let KalkValue::Matrix(rows) = x {
+                let mut new_rows = Vec::new();
+                let mut success = true;
+                for row in rows {
+                    let mut new_row = Vec::new();
+                    for value in row {
+                        if let Some(result) = prelude::call_unary_func(
+                            context,
+                            &identifier.full_name,
+                            value,
+                            &context.angle_unit.clone(),
+                        ) {
+                            new_row.push(result.0?);
+                        } else {
+                            success = false;
+                            break;
+                        }
+                    }
 
-            for n in start..=end {
-                context.sum_n_value = Some(n);
-                let eval = eval_expr(context, &expressions[2], "")?;
-                if sum_else_prod {
-                    sum.value += eval.value;
-                    sum.imaginary_value += eval.imaginary_value;
-                } else {
-                    sum.value *= eval.value;
-                    sum.imaginary_value *= eval.imaginary_value;
+                    new_rows.push(new_row);
+                }
+
+                if success {
+                    return Ok(KalkValue::Matrix(new_rows));
                 }
             }
-
-            context.sum_n_value = None;
-            sum.unit = unit.into();
-
-            return Ok(sum);
         }
-        "integrate" | "integral" | "∫" => {
-            // Make sure either 3 or 4 arguments were supplied.
-            if expressions.len() < 3 || expressions.len() > 4 {}
 
-            return match expressions.len() {
-                3 => calculus::integrate_with_unknown_variable(
-                    context,
-                    &expressions[0],
-                    &expressions[1],
-                    &expressions[2],
-                ),
-                4 => calculus::integrate(
-                    context,
-                    &expressions[0],
-                    &expressions[1],
-                    &expressions[2],
-                    if let Expr::Var(integration_variable) = &expressions[3] {
-                        &integration_variable.full_name[1..]
-                    } else {
-                        return Err(CalcError::ExpectedDx);
-                    },
-                ),
-                _ => Err(CalcError::IncorrectAmountOfArguments(
-                    3,
-                    "integrate".into(),
-                    expressions.len(),
-                )),
-            };
-        }
-        _ => (),
+        return result;
     }
 
     // Symbol Table
@@ -390,7 +526,7 @@ pub(crate) fn eval_fn_call_expr(
     match stmt_definition {
         Some(Stmt::FnDecl(_, arguments, fn_body)) => {
             if arguments.len() != expressions.len() {
-                return Err(CalcError::IncorrectAmountOfArguments(
+                return Err(KalkError::IncorrectAmountOfArguments(
                     arguments.len(),
                     identifier.full_name.clone(),
                     expressions.len(),
@@ -400,7 +536,7 @@ pub(crate) fn eval_fn_call_expr(
             // Initialise the arguments as their own variables.
             let mut new_argument_values = Vec::new();
             for (i, argument) in arguments.iter().enumerate() {
-                let argument_identifier = if argument.contains("-") {
+                let argument_identifier = if argument.contains('-') {
                     let identifier_parts: Vec<&str> = argument.split('-').collect();
                     Identifier::parameter_from_name(identifier_parts[1], identifier_parts[0])
                 } else {
@@ -411,11 +547,11 @@ pub(crate) fn eval_fn_call_expr(
                     Box::new(crate::ast::build_literal_ast(&eval_expr(
                         context,
                         &expressions[i],
-                        "",
+                        None,
                     )?)),
                 );
 
-                // Don't set these values just yet, since
+                // Don't set these values just yet,
                 // to avoid affecting the value of arguments
                 // during recursion.
                 new_argument_values.push((argument, var_decl));
@@ -437,32 +573,268 @@ pub(crate) fn eval_fn_call_expr(
             let fn_value = eval_expr(context, &fn_body, unit);
 
             // Revert to original argument values
-            for old_argument_value in old_argument_values {
-                if let Some(old_argument_value) = old_argument_value {
-                    context.symbol_table.insert(old_argument_value);
-                }
+            for old_argument_value in old_argument_values.into_iter().flatten() {
+                context.symbol_table.insert(old_argument_value);
             }
 
             fn_value
         }
-        _ => Err(CalcError::UndefinedFn(identifier.full_name.clone())),
+        _ => Err(KalkError::UndefinedFn(identifier.full_name.clone())),
     }
 }
 
+fn eval_loop(
+    context: &mut Context,
+    identifier: &Identifier,
+    var_name: &str,
+    start_expr: &Expr,
+    end_expr: &Expr,
+    expression: &Expr,
+    unit: Option<String>,
+) -> Result<KalkValue, KalkError> {
+    if context.sum_variables.is_none() {
+        context.sum_variables = Some(Vec::new());
+    }
+
+    {
+        let sum_variables = context.sum_variables.as_mut().unwrap();
+        sum_variables.push(SumVar {
+            name: var_name.into(),
+            value: 0,
+        });
+    }
+
+    let start = eval_expr(context, start_expr, None)?.to_f64() as i128;
+    let end = eval_expr(context, end_expr, None)?.to_f64() as i128;
+    let sum_else_prod = match identifier.full_name.as_ref() {
+        "sum" => true,
+        "prod" => false,
+        _ => unreachable!(),
+    };
+    let mut sum = if sum_else_prod {
+        KalkValue::from(0f64)
+    } else {
+        KalkValue::from(1f64)
+    };
+
+    for n in start..=end {
+        let sum_variables = context.sum_variables.as_mut().unwrap();
+        sum_variables.last_mut().unwrap().value = n;
+
+        let eval = eval_expr(context, expression, None)?;
+        if sum_else_prod {
+            sum = sum.add(context, eval)?;
+        } else {
+            sum = sum.mul(context, eval)?;
+        }
+    }
+
+    let sum_variables = context.sum_variables.as_mut().unwrap();
+    sum_variables.pop();
+
+    let (sum_real, sum_imaginary, _) = as_number_or_zero!(sum);
+
+    Ok(KalkValue::Number(sum_real, sum_imaginary, unit))
+}
+
 fn eval_piecewise(
     context: &mut Context,
-    pieces: &Vec<crate::ast::ConditionalPiece>,
-    unit: &str,
-) -> Result<KalkNum, CalcError> {
+    pieces: &[crate::ast::ConditionalPiece],
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
     for piece in pieces {
-        if let Some(condition_is_true) = eval_expr(context, &piece.condition, unit)?.boolean_value {
+        if let KalkValue::Boolean(condition_is_true) = eval_expr(context, &piece.condition, unit)? {
             if condition_is_true {
-                return Ok(eval_expr(context, &piece.expr, unit)?);
+                return eval_expr(context, &piece.expr, unit);
             }
         }
     }
 
-    Err(CalcError::PiecewiseConditionsAreFalse)
+    Err(KalkError::PiecewiseConditionsAreFalse)
+}
+
+fn eval_vector(context: &mut Context, values: &[Expr]) -> Result<KalkValue, KalkError> {
+    let mut eval_values = Vec::new();
+    for value in values {
+        eval_values.push(eval_expr(context, value, None)?);
+    }
+
+    Ok(KalkValue::Vector(eval_values))
+}
+
+fn eval_matrix(context: &mut Context, rows: &[Vec<Expr>]) -> Result<KalkValue, KalkError> {
+    let mut eval_rows = Vec::new();
+    for row in rows {
+        let mut eval_row = Vec::new();
+        for value in row {
+            eval_row.push(eval_expr(context, value, None)?)
+        }
+
+        eval_rows.push(eval_row);
+    }
+
+    Ok(KalkValue::Matrix(eval_rows))
+}
+
+fn eval_indexer(
+    context: &mut Context,
+    var: &Expr,
+    index_expressions: &[Expr],
+    unit: Option<&String>,
+) -> Result<KalkValue, KalkError> {
+    let var_value = eval_expr(context, var, unit)?;
+    match var_value {
+        KalkValue::Vector(values) => {
+            if index_expressions.len() != 1 {
+                return Err(KalkError::IncorrectAmountOfIndexes(
+                    index_expressions.len(),
+                    1,
+                ));
+            }
+
+            let index = as_indices(context, index_expressions)?[0];
+            if let Some(value) = values.get(index - 1) {
+                Ok(value.clone())
+            } else {
+                Err(KalkError::ItemOfIndexDoesNotExist(vec![index]))
+            }
+        }
+        KalkValue::Matrix(rows) => {
+            let indices = as_indices(context, index_expressions)?;
+            if indices.len() == 1 {
+                let row_index = indices[0];
+                return if let Some(row) = rows.get(row_index - 1) {
+                    Ok(KalkValue::Vector(row.clone()))
+                } else {
+                    Err(KalkError::ItemOfIndexDoesNotExist(vec![row_index]))
+                };
+            }
+
+            if indices.len() != 2 {
+                return Err(KalkError::IncorrectAmountOfIndexes(indices.len(), 2));
+            }
+
+            let row_index = indices[0];
+            let column_index = indices[1];
+            if row_index == 0 || column_index == 0 {
+                return Err(KalkError::ItemOfIndexDoesNotExist(vec![
+                    row_index,
+                    column_index,
+                ]));
+            }
+
+            if let Some(row) = rows.get(row_index - 1) {
+                if let Some(value) = row.get(column_index - 1) {
+                    return Ok(value.clone());
+                }
+            }
+
+            Err(KalkError::ItemOfIndexDoesNotExist(vec![
+                row_index,
+                column_index,
+            ]))
+        }
+        _ => Err(KalkError::CanOnlyIndexX),
+    }
+}
+
+fn as_indices(context: &mut Context, expressions: &[Expr]) -> Result<Vec<usize>, KalkError> {
+    let mut indices = Vec::new();
+    for expr in expressions {
+        let value = eval_expr(context, expr, None)?;
+        if value.has_imaginary() {
+            return Err(KalkError::CannotIndexByImaginary);
+        }
+
+        let index = value.to_f64() as usize;
+        if index == 0 {
+            return Err(KalkError::ItemOfIndexDoesNotExist(vec![index]));
+        }
+
+        indices.push(index);
+    }
+
+    Ok(indices)
+}
+
+fn eval_comprehension(
+    context: &mut Context,
+    left: &Expr,
+    conditions: &[Expr],
+    vars: &[RangedVar],
+) -> Result<Vec<KalkValue>, KalkError> {
+    if vars.len() != conditions.len() {
+        return Err(KalkError::InvalidComprehension(String::from("Expected a new variable to be introduced for every condition (conditions are comma separated).")));
+    }
+
+    let condition = conditions.first().unwrap();
+    let var = vars.first().unwrap();
+    context.symbol_table.insert(Stmt::VarDecl(
+        Identifier::from_full_name(&var.name),
+        Box::new(Expr::Literal(0f64)),
+    ));
+
+    let min = eval_expr(context, &var.min, None)?.to_f64() as i32;
+    let max = eval_expr(context, &var.max, None)?.to_f64() as i32;
+
+    let mut values = Vec::new();
+    for i in min..max {
+        context.symbol_table.set(Stmt::VarDecl(
+            Identifier::from_full_name(&var.name),
+            Box::new(Expr::Literal(i as f64)),
+        ));
+
+        if conditions.len() > 1 {
+            let x = eval_comprehension(context, left, &conditions[1..], &vars[1..]);
+            if let Err(err) = x {
+                context.symbol_table.get_and_remove_var(&var.name);
+                return Err(err);
+            }
+
+            for value in x.unwrap() {
+                values.push(Ok(value));
+            }
+        }
+
+        let condition = eval_expr(context, condition, None);
+        match condition {
+            Ok(KalkValue::Boolean(boolean)) => {
+                if boolean && vars.len() == 1 {
+                    values.push(eval_expr(context, left, None));
+                }
+            }
+            Err(err) => values.push(Err(err)),
+            _ => (),
+        }
+    }
+
+    context.symbol_table.get_and_remove_var(&var.name);
+
+    let mut unwrapped_values = Vec::new();
+    for value in values {
+        match value {
+            Ok(unwrapped_value) => unwrapped_values.push(unwrapped_value),
+            Err(err) => return Err(err),
+        }
+    }
+
+    Ok(unwrapped_values)
+}
+
+fn eval_equation(
+    context: &mut Context,
+    left: &Expr,
+    right: &Expr,
+    unknown_var: &Identifier,
+) -> Result<KalkValue, KalkError> {
+    context.is_approximation = true;
+
+    let expr = Expr::Binary(
+        Box::new(left.clone()),
+        TokenKind::Minus,
+        Box::new(right.clone()),
+    );
+    numerical::find_root(context, &expr, &unknown_var.full_name)
 }
 
 #[cfg(test)]
@@ -497,7 +869,7 @@ mod tests {
         );
     }
 
-    fn interpret_with_unit(stmt: Stmt) -> Result<Option<KalkNum>, CalcError> {
+    fn interpret_with_unit(stmt: Stmt) -> Result<Option<CalculationResult>, KalkError> {
         let mut symbol_table = SymbolTable::new();
         symbol_table
             .insert(DEG_RAD_UNIT.clone())
@@ -506,9 +878,9 @@ mod tests {
         context(&mut symbol_table, "rad").interpret(vec![stmt])
     }
 
-    fn interpret(stmt: Stmt) -> Result<Option<KalkNum>, CalcError> {
+    fn interpret(stmt: Stmt) -> Result<Option<KalkValue>, KalkError> {
         if let Some(result) = interpret_with_unit(stmt)? {
-            Ok(Some(result))
+            Ok(Some(result.get_value()))
         } else {
             Ok(None)
         }
@@ -524,10 +896,18 @@ mod tests {
         Context::new(symbol_table, angle_unit, None)
     }
 
-    fn cmp(x: KalkNum, y: f64) -> bool {
+    fn cmp(x: KalkValue, y: f64) -> bool {
         (x.to_f64() - y).abs() < 0.0001
     }
 
+    fn bool(x: &KalkValue) -> bool {
+        if let KalkValue::Boolean(boolean_value) = x {
+            *boolean_value
+        } else {
+            false
+        }
+    }
+
     #[test]
     fn test_literal() {
         let stmt = Stmt::Expr(literal(1f64));
@@ -556,40 +936,28 @@ mod tests {
         assert_eq!(interpret(pow).unwrap().unwrap().to_f64(), 8f64);
 
         let result = interpret(equals).unwrap().unwrap();
-        assert_eq!(
-            (result.to_f64(), result.boolean_value.unwrap()),
-            (3f64, false)
-        );
+        assert!(!bool(&result));
+        assert!(result.to_f64().is_nan());
 
         let result = interpret(not_equals).unwrap().unwrap();
-        assert_eq!(
-            (result.to_f64(), result.boolean_value.unwrap()),
-            (3f64, true)
-        );
+        assert!(bool(&result));
+        assert!(result.to_f64().is_nan());
 
         let result = interpret(greater_than).unwrap().unwrap();
-        assert_eq!(
-            (result.to_f64(), result.boolean_value.unwrap()),
-            (3f64, false)
-        );
+        assert!(!bool(&result));
+        assert!(result.to_f64().is_nan());
 
         let result = interpret(less_than).unwrap().unwrap();
-        assert_eq!(
-            (result.to_f64(), result.boolean_value.unwrap()),
-            (3f64, true)
-        );
+        assert!(bool(&result));
+        assert!(result.to_f64().is_nan());
 
         let result = interpret(greater_or_equals).unwrap().unwrap();
-        assert_eq!(
-            (result.to_f64(), result.boolean_value.unwrap()),
-            (3f64, false)
-        );
+        assert!(!bool(&result));
+        assert!(result.to_f64().is_nan());
 
         let result = interpret(less_or_equals).unwrap().unwrap();
-        assert_eq!(
-            (result.to_f64(), result.boolean_value.unwrap()),
-            (3f64, true)
-        );
+        assert!(bool(&result));
+        assert!(result.to_f64().is_nan());
     }
 
     #[test]
@@ -636,11 +1004,16 @@ mod tests {
             rad_context
                 .interpret(vec![implicit.clone()])
                 .unwrap()
-                .unwrap(),
+                .unwrap()
+                .get_value(),
             0.84147098
         ));
         assert!(cmp(
-            deg_context.interpret(vec![implicit]).unwrap().unwrap(),
+            deg_context
+                .interpret(vec![implicit])
+                .unwrap()
+                .unwrap()
+                .get_value(),
             0.01745240
         ));
     }
@@ -666,7 +1039,7 @@ mod tests {
 
         assert_eq!(
             interpret(stmt),
-            Err(CalcError::UndefinedVar(String::from("x")))
+            Err(KalkError::UndefinedVar(String::from("x")))
         );
     }
 
@@ -706,7 +1079,7 @@ mod tests {
 
         assert_eq!(
             interpret(stmt),
-            Err(CalcError::UndefinedFn(String::from("f")))
+            Err(KalkError::UndefinedFn(String::from("f")))
         );
     }
 
@@ -716,7 +1089,7 @@ mod tests {
         let stmt = Stmt::Expr(fn_call(
             "sum",
             vec![
-                *literal(start),
+                *binary(var("n"), TokenKind::Equals, literal(start)),
                 *literal(to),
                 *binary(var("n"), TokenKind::Plus, literal(3f64)),
             ],
